@@ -8,7 +8,8 @@
 use lib '../lib';
 use AnyEvent::STOMP::Client;
 
-my $stomp_client = AnyEvent::STOMP::Client->connect();
+my $stomp_client = new AnyEvent::STOMP::Client()
+$stomp_client->connect();
 
 $stomp_client->on_connected(
     sub {
